##############################################################################
#
#    Accounting periods, for Odoo
#    Copyright (C) 2018 XCG Consulting <http://odoo.consulting>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import datetime

from odoo import _, api, exceptions, fields, models


class AccountMove(models.Model):
    """Add a period & dates onto accounting documents.
    """

    _inherit = "account.move"

    accounting_date = fields.Date(
        string="Accounting date",
        copy=False,
        help="Validation date of the accounting document.",
        states={"posted": [("readonly", True), ("required", True)]},
        required=False,
        readonly=False,
    )

    period_id = fields.Many2one(
        comodel_name="account.period",
        string="Period",
        ondelete="restrict",
        help="The period this accounting document is in.",
        states={"posted": [("readonly", True), ("required", True)]},
        required=False,
        readonly=False,
    )

    transaction_date = fields.Date(
        string="Transaction date",
        copy=False,
        help=(
            "Invoicing date when provided; otherwise this is the accounting "
            "date."
        ),
        states={"posted": [("readonly", True)]},
    )

    @api.model_create_multi
    def create(self, vals_list):
        """Override to set a transaction date from invoices.
        """

        for vals in vals_list:
            invoice_date = vals.get("invoice_date")
            if invoice_date:
                vals["transaction_date"] = invoice_date

        return super(AccountMove, self).create(vals_list)

    def post(self):
        """Override accounting document validation to fill accounting dates &
        period.
        """

        self.check_accounting_date_integrity()
        self.fill_accounting_dates()

        return super(AccountMove, self).post()

    def check_accounting_date_integrity(self):

        if any(
            self.mapped(
                lambda r: (
                    r.accounting_date > fields.Date.today()
                    if r.accounting_date
                    else False
                )
            )
        ):

            raise exceptions.ValidationError(
                _(
                    "The accounting date must be anterior to the validation "
                    "date or the same."
                )
            )

        return True

    def check_period_integrity(self):

        if any(
            self.mapped(lambda r: r.period_id.date_start > fields.Date.today())
        ):

            raise exceptions.ValidationError(
                _(
                    "The accounting period must match the validation "
                    "date or be anterior to it."
                )
            )

        return True

    def fill_accounting_dates(self):
        """- Set the accounting date.
        - Also set the transaction date ("transaction_date" field) when empty.
        - Force the period to always be around the current date.
        - Only select open periods.
        """

        today = datetime.date.today()

        for accdoc in self:

            acc_date = accdoc.accounting_date or today

            # Cache some data.
            company = accdoc.company_id

            # Set the acc_date only if the force_period_on_date
            # context has provided.
            if self.env.context.get("force_period_on_date"):
                # If accounting document date is empty, get today date.
                acc_date = accdoc.date or acc_date

            if not accdoc.period_id:
                # Periods are ordered by date so selecting the first one is
                # fine.
                period = self.env["account.period"].search(
                    [
                        ("company_id", "=", company.id),
                        ("date_start", "<=", acc_date),
                        ("date_effective_cutoff", ">=", acc_date),
                        ("state", "!=", "done"),
                    ],
                    limit=1,
                )
                if not period:
                    raise exceptions.Warning(
                        _('No period found around %s in the "%s" company.')
                        % (acc_date, company.sudo().name)
                    )

                accdoc_values = {"period_id": period.id}
            else:
                accdoc.check_period_integrity()
                period = accdoc.period_id
                accdoc_values = {}

            # When we are between the period end and cut-off date, force the
            # last day of the period.
            in_cutoff = False
            if accdoc.journal_id.type == "sale":
                period_end = period.date_stop
                in_cutoff = acc_date > period_end
                acc_date = period_end if in_cutoff else acc_date

            # The data to update the accounting document with.
            accdoc_values.update({"accounting_date": acc_date})

            # Set a transaction date when no previous one set. Also, force it
            # during cut-off.
            if in_cutoff or not accdoc.transaction_date:
                accdoc_values["transaction_date"] = acc_date

            # Ready! Update the accounting document.
            accdoc.write(accdoc_values)

        return True
